from functools import partial
from inspect import iscoroutinefunction as is_async
from typing import (
    get_args, get_origin,
    Union, Annotated, Any, Type, Callable, Iterable, List, Mapping, ForwardRef
)

from pydantic import validate_arguments
import orjson as json

from pyservices.signals import service_signals, endpoint_signals


class _Dependency:
    """
    """

    new: bool = False
    new_recursive: bool = False
    args: Union[Iterable, None] = None
    kwargs: Union[Mapping, None] = None

    def mutate(
        self,
        new: bool = None,
        new_recursive: bool = None,
        args: Union[Iterable, None] = None,
        kwargs: Union[Mapping, None] = None
    ) -> '_Dependency':
        mutated = _Dependency()
        mutated.new = self.new if new is None else new
        mutated.new_recursive = self.new_recursive if new_recursive is None else new_recursive
        mutated.args = self.args if args is None else args
        mutated.kwargs = self.kwargs if kwargs is None else kwargs
        return mutated

    __call__ = mutate

    def __str__(
        self
    ) -> str:
        output = 'Depend'
        if self.new:
            output += ' new'
        if self.new_recursive:
            output += ' new_recursive'
        if self.args:
            output += ' args'
        if self.kwargs:
            output += ' kwargs'
        return output


Depend = _Dependency()


class _ServiceEndpoint:

    service_class: Type
    method: Callable
    _async_method: bool
    _json_output: bool

    def __init__(
        self,
        service_class: Type,
        method: Callable,
        validate_payload: bool = True,
        json_output: bool = True,
    ):
        self.service_class = service_class
        self._async_method = is_async(method)
        if validate_payload:
            self.method = validate_arguments(method)
        else:
            self.method = method
        self._json_output = json_output

    async def call(
        self,
        *args: Iterable,
        **kwargs: Mapping
    ) -> Any:
        """
        """
        await service_signals.fire('begin')
        try:
            instance = await initialize(self.service_class)
        except Exception as e:
            await service_signals.fire('catch', e)
            raise e
        await service_signals.fire('end')
        await endpoint_signals.fire('begin')
        try:
            if self._async_method:
                output = await self.method(instance, *args, **kwargs)
            else:
                output = self.method(instance, *args, **kwargs)
        except Exception as e:
            await endpoint_signals.fire('catch', e)
            raise e
        await endpoint_signals.fire('end')
        if self._json_output:
            return json.dumps(output)
        return output

    __call__ = call


def get_class_methods(
    klass
) -> List[Callable]:
    """
    Return user defined class methods as list
    """
    class_attributes = [getattr(klass, a) for a in dir(klass) if a.startswith('_') is False]
    return [m for m in class_attributes if callable(m)]


class ServiceNotFound(Exception):
    """
    """


def hash_class(
    klass: Union[Type, ForwardRef, str]
) -> str:
    if isinstance(klass, ForwardRef) and klass.__forward_is_class__:
        return klass.__forward_arg__
    if isinstance(klass, str):
        return klass
    return klass.__name__


class __ServiceRegistry:

    __registry = {}

    def get(
        self,
        service_annotation: Union[Type, ForwardRef, str]
    ) -> Type:
        """
        Raises:
            - ServiceNotFound
        """
        service_hash = hash_class(service_annotation)
        service_class = self.__registry.get(service_hash, None)
        if service_class is None:
            raise ServiceNotFound()
        return service_class

    def set(
        self,
        klass
    ) -> None:
        """
        """
        service_hash = hash_class(klass)
        self.__registry[service_hash] = klass


service_registry = __ServiceRegistry()


def make_service(
    klass
):
    service_registry.set(klass)
    for method in get_class_methods(klass):
        method.as_endpoint = partial(_ServiceEndpoint, klass, method)


def service(
    klass
):
    def wrap(klass):
        make_service(klass)
        return klass
    if klass is None:
        return wrap
    return wrap(klass)


async def initialize(
    service_class: Union[ForwardRef, str, Type],
    __already_initialized: Mapping = None,
    __path: List[Type] = None,
):
    """
    """
    if __already_initialized is None:
        __already_initialized = {}

    # Creates new list of service classes
    # if __path is None:
    #     __path = [service_class]
    # else:
    #     __path = __path + [service_class]

    # initializes class
    service_class = service_registry.get(service_class)
    root_instance = service_class()
    root_class_hash = hash_class(service_class)
    __already_initialized[root_class_hash] = root_instance

    if not hasattr(root_instance, '__annotations__'):
        # set default attribute __annotations__
        root_instance.__annotations__ = {}

    for child_service_name, v in root_instance.__annotations__.items():
        if get_origin(v) is not Annotated:
            continue

        try:
            child_class, dependency = get_args(v)
        except Exception:
            continue

        if not isinstance(dependency, _Dependency):
            # annotated value is not instance of _Dependency
            continue

        child_class_hash = hash_class(child_class)

        if dependency.new:
            # initialize (again) only this dependency
            child_instance = None
            child_already_initialized = __already_initialized
        elif dependency.new_recursive:
            # reset initialized instances
            child_instance = None
            child_already_initialized = {}
        else:
            # get already initialized instance
            child_instance = __already_initialized.get(child_class_hash, None)
            child_already_initialized = __already_initialized

        if child_instance is None:
            # otherwise initialize it recursively
            child_instance = await initialize(child_class, child_already_initialized, __path)
            child_already_initialized[child_class_hash] = child_instance

        # set child service as attribute
        setattr(root_instance, child_service_name, child_instance)

    # print(__path, 'initialized')
    return root_instance




import asyncio
from inspect import iscoroutinefunction as is_async
from typing import Mapping, List, Callable, Union


class SignalManager:

    __signals: Mapping[str, List[Callable]] = {}

    def add(
        self,
        signal_name: str,
        procedure: Callable
    ) -> None:
        assert callable(procedure), 'procedure must be Callable'
        self.__signals.setdefault(signal_name, [])
        self.__signals[signal_name].append(procedure)

    def on(
        self,
        signal_name: Union[str, Callable] = None
    ) -> Callable:
        """
        Uses procedure name if signal_name is not passed
        Returns passed procedure
        >>> @signals.on
        >>> async def ():
        """
        def decorate(procedure: Callable):
            _signal_name = procedure.__name__ if signal_name is None else signal_name
            self.add(_signal_name, procedure)
            return procedure
        if callable(signal_name):
            _signal_name, procedure = signal_name.__name__, signal_name
            self.add(_signal_name, procedure)
            return procedure
        return decorate

    async def fire(
        self,
        signal_name: str,
        *args,
        **kwargs
    ) -> None:
        P = self.__signals.get(signal_name, [])
        C = []
        for procedure in P:
            if is_async(procedure):
                c = procedure(*args, **kwargs)
                C.append(c)
            else:
                procedure(*args, **kwargs)
        await asyncio.gather(*C)


service_signals = SignalManager()
endpoint_signals = SignalManager()




from setuptools import setup


with open('./README.md', 'r', encoding='utf-8') as file:
    long_description = file.read()


setup(
    name='pyservices',
    version='0.0.1.1',
    author='Aidar Turkenov',
    author_email='a.k.turken0v@gmail.com',
    description='',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/aturkenov/pyservices',
    project_urls={
        'Bug Tracker': 'https://github.com/aturkenov/pyservices/issues',
        'Discussions': 'https://github.com/aturkenov/pyservices/discussions',
    },
    classifiers=[
        'Typing :: Typed',
        'Framework :: AsyncIO',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Programming Language :: Python :: 3.11',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    packages=['pyservices'],
    install_requires=[
        'orjson',
        'pydantic',
    ],
    python_requires='>=3.6',
)

